/**********************************************************
* Filename:    FSM_DD.h
*
* Description: OLT Data Detector state machine
*
*********************************************************/

#ifndef _FSM_DD_H_INCLUDED_
#define _FSM_DD_H_INCLUDED_

#include "_queue.h"
#include "FSM_base.h"

const int32s FIFO_DD_OLT_SIZE = BLK_ROUNDUP( FEC_PAYLOAD_BYTES + MAX_FRAME_BYTES, FEC_PAYLOAD_BYTES ) * FEC_PSIZE + 1;
const int32s FIFO_DD_ONU_SIZE = DELAY_BOUND;
const int32s FIFO_EXIT_SIZE   = 1000; //FIFO_DD_OLT_SIZE;

/*
#define SP              _66b_t( 0, N_BLOCK )
#define PARITY_BLOCK    _66b_t( 0, P_BLOCK )
#define IDLE_BLOCK      _66b_t( 0, C_BLOCK )
#define ZERO_BLOCK      _66b_t( 0, Z_BLOCK )
#define BURST_DELIMITER _66b_t( 0, L_BLOCK )
*/
#define PARITY_BLOCK    _66b_t( 0, P_BLOCK )

#define SP              _66b_t( 0, C_BLOCK )
#define IDLE_BLOCK      _66b_t( 0, C_BLOCK )
#define ZERO_BLOCK      _66b_t( 0, C_BLOCK )
#define BURST_DELIMITER _66b_t( 0, C_BLOCK )

/////////////////////////////////////////////////////////////
// OLT DATA DETECTOR
/////////////////////////////////////////////////////////////
class fsm_olt_data_detector_t: public fsm_base_t< DLY_DATA_DET, _66b_t >
{
protected:
    int32s  protected_block_count;
	Queue< _66b_t, FIFO_DD_OLT_SIZE > FIFO_DD;

	/////////////////////////////////////////////////////////////
    // FIFO access wrapper function
    /////////////////////////////////////////////////////////////
	void  AddToFIFO( _66b_t vector )
    {
		if( FIFO_DD.IsFull() )
			MSG_WARN( "Attempt to receive into a full FIFO_DD. Block is discarded." );
		
		FIFO_DD.Add( vector );
	}

	/////////////////////////////////////////////////////////////
    // FIFO access wrapper function
    /////////////////////////////////////////////////////////////
	_66b_t SendFromFIFO( void )
    {
		if( FIFO_DD.IsEmpty() )
		{
			MSG_WARN( "Attempt to transmit from an empty FIFO_DD" );
			return _66b_t();
		}
		return FIFO_DD.Get();
	}

    /////////////////////////////////////////////////////////////
    // 
    /////////////////////////////////////////////////////////////
    void RepeatBlocks( _66b_t vector, int16s number )
    {
        while( number-- > 0 )
	        AddToFIFO( vector );
    }

    /////////////////////////////////////////////////////////////
    // 
    /////////////////////////////////////////////////////////////
    void ReceiveUnit( _66b_t vector )
    {
        protected_block_count++;
		AddToFIFO( vector );
		
        if( protected_block_count == FEC_DSIZE )
        {
            RepeatBlocks( _66b_t(0, P_BLOCK, -2), FEC_PSIZE );
            protected_block_count = 0;
		}
	}

	/////////////////////////////////////////////////////////////
    // 
    /////////////////////////////////////////////////////////////
	inline _66b_t TransmitUnit() { return SendFromFIFO(); }
    
public:

    fsm_olt_data_detector_t()
    {
        protected_block_count = 0;
    }
};




/////////////////////////////////////////////////////////////
// ONU DATA DETECTOR
/////////////////////////////////////////////////////////////
class fsm_onu_data_detector_t: public fsm_base_t< DLY_DATA_DET, _66b_t >
{
    enum state_t 
    {
        STATE_LASER_IS_OFF = 1,
        STATE_TX_BURST_PRE = 2,
        STATE_TX_DELIMITER = 3,
        STATE_FEC_IS_ON    = 4
    };

private:
    int32s  unprotected_block_count;
    int32s  protected_block_count;
    int32s  idle_block_count;
    int32s  sync_block_count;
    state_t current_state;

    Queue< _66b_t, FIFO_DD_ONU_SIZE >   FIFO_DD;
    Queue< _66b_t, FIFO_EXIT_SIZE >     FIFO_EXIT;

    /////////////////////////////////////////////////////////////
    // void RemoveHeadBlock( void )
    /////////////////////////////////////////////////////////////
	void RemoveHeadBlock( int16s number = 1 )
    {
        while( number-- > 0 && !FIFO_DD.IsEmpty() )
	        FIFO_DD.Get();

        if( number > 0 )
            MSG_WARN( "Attempt to remove block from an empty FIFO_DD" );
	}

    /////////////////////////////////////////////////////////////
    // void ReceiveNextBlock( _66b_t vector )
    /////////////////////////////////////////////////////////////
	inline void ReceiveNextBlock( _66b_t vector )
    {
        RemoveHeadBlock();
		FIFO_DD.Add( vector );  // add tail to FIFO_DD
	}

    /////////////////////////////////////////////////////////////
    // 
    /////////////////////////////////////////////////////////////
    void TransmitBlock( _66b_t vector, int16s number = 1 )
    {
        while( number-- > 0 && !FIFO_EXIT.IsFull() )
	        FIFO_EXIT.Add( vector );
       
        if( number > 0 )
            MSG_WARN( "Attempt to write into a full FIFO_EXIT. Block(s) discarded." );
    }

    /////////////////////////////////////////////////////////////
    // processing done in state LASER_IS_OFF
    /////////////////////////////////////////////////////////////
    void StateLaserIsOff( _66b_t vector )
    {
        MSG_INFO("Laser Is Off");
        TransmitBlock( SP );
        unprotected_block_count++;
		
        if( unprotected_block_count >= FEC_DSIZE )
        {
            TransmitBlock( SP, FEC_PSIZE );
            unprotected_block_count -= FEC_DSIZE;
		}

        ReceiveNextBlock( vector );

        if( vector.SyncHeader() == SH_DATA )
        {
            sync_block_count = 0;
            // turn laser on
            // PMD_SIGNAL.request( true );
            current_state = STATE_TX_BURST_PRE;
        }
	}

    /////////////////////////////////////////////////////////////
    // processing done in state TRANSMIT_BURST_PREAMBLE
    /////////////////////////////////////////////////////////////
    void StateTxBurstPreamble( _66b_t vector )
    {
        MSG_INFO("TxBurstPreamble");
        TransmitBlock( SP );
        unprotected_block_count++;
		sync_block_count ++;
		
        if( unprotected_block_count >= FEC_DSIZE )
        {
            TransmitBlock( SP, FEC_PSIZE );
            unprotected_block_count -= FEC_DSIZE;
		}

        ReceiveNextBlock( vector );

        idle_block_count = (vector.SyncHeader() == SH_CTRL)? idle_block_count + 1 : -1;
        
        if( sync_block_count == SYNC_LENGTH )
            current_state = STATE_TX_DELIMITER;
	}

    /////////////////////////////////////////////////////////////
    // processing done in state TRANSMIT_BURST_DELIMITER
    /////////////////////////////////////////////////////////////
    void StateTxDelimiter( _66b_t vector )
    {
        MSG_INFO("TxDelimiter");
        TransmitBlock( BURST_DELIMITER );
        unprotected_block_count++;
        ReceiveNextBlock( vector );
        current_state = STATE_FEC_IS_ON;
    }

    /////////////////////////////////////////////////////////////
    // processing done in state FEC_IS_ON
    /////////////////////////////////////////////////////////////
    void StateFecIsOn( _66b_t vector )
    {
        // pass block to FEC encoder
        // FEC_encoder( FIFO_DD.PeekHead() );
        MSG_INFO("Fec Is On");
        TransmitBlock( FIFO_DD.Peek() );
        protected_block_count++;
		
        if( protected_block_count == FEC_DSIZE )
        {
            TransmitBlock( PARITY_BLOCK, FEC_PSIZE );
            protected_block_count = 0;
		}

        ReceiveNextBlock( vector );

        idle_block_count = (vector.SyncHeader() == SH_CTRL)? idle_block_count + 1 : -1;
        
        if( idle_block_count >= DELAY_BOUND && protected_block_count == 0 )
        {
            // turn laser on
            // PMD_SIGNAL.request( false );

            // send zeroes
            TransmitBlock( ZERO_BLOCK, 3 );
            // remove corresponding number of blocks from the FIFO
            RemoveHeadBlock( 3 );
            unprotected_block_count += 3;

            current_state = STATE_LASER_IS_OFF;
        }
       
	} 

    /////////////////////////////////////////////////////////////
    // 
    /////////////////////////////////////////////////////////////
    void ReceiveUnit( _66b_t vector )
    {
        //MSG_INFO( "====================== rx " << (int32s)current_state << "  " << FIFO_EXIT.GetSize() <<":" << FIFO_DD.GetSize() << "=== vect " << BlockName( vector.T_TYPE()) );
        switch( current_state )
        {
            case STATE_LASER_IS_OFF:  StateLaserIsOff( vector );        break;
            case STATE_TX_BURST_PRE:  StateTxBurstPreamble( vector );   break;
            case STATE_TX_DELIMITER:  StateTxDelimiter( vector );       break;
            case STATE_FEC_IS_ON:     StateFecIsOn( vector );           break;
            default:                  MSG_WARN( "ONU DATA DETECTOR is in unknown state" );
        };
    }

    /////////////////////////////////////////////////////////////
    // 
    /////////////////////////////////////////////////////////////
	inline _66b_t TransmitUnit() 
    { 
        if( !FIFO_EXIT.IsEmpty() )
        {
            MSG_INFO( "=== ex tx " << (int32s)current_state << "  " << FIFO_EXIT.GetSize() -1 <<":" << FIFO_DD.GetSize()); 
            MSG_INFO( "    == col[0] " << BlockName( FIFO_EXIT.Peek()[0].C_TYPE()) << " #" << FIFO_EXIT.Peek()[0].GetSeqNumber() );
            MSG_INFO( "    == col[1] " << BlockName( FIFO_EXIT.Peek()[1].C_TYPE()) << " #" << FIFO_EXIT.Peek()[1].GetSeqNumber() );
            return FIFO_EXIT.Get();
        }

        if( !FIFO_DD.IsEmpty() )
        {
            MSG_INFO( "=== dd tx " << (int32s)current_state << "  " << FIFO_EXIT.GetSize() <<":" << FIFO_DD.GetSize()); 
            MSG_INFO( "    == col[0] " << BlockName( FIFO_DD.Peek()[0].C_TYPE()) << " #" << FIFO_DD.Peek()[0].GetSeqNumber() );
            MSG_INFO( "    == col[1] " << BlockName( FIFO_DD.Peek()[1].C_TYPE()) << " #" << FIFO_DD.Peek()[1].GetSeqNumber() );
            return FIFO_DD.Get();
        }

        MSG_WARN( "Attempt to transmit from an empty FIFO_DD" );
        return IDLE_BLOCK;
    }

public:

    fsm_onu_data_detector_t() 
    {
        unprotected_block_count = 0;
        protected_block_count = 0;
        idle_block_count = -1;
        sync_block_count = 0;
        current_state = STATE_LASER_IS_OFF;

        while( FIFO_DD.GetSize() < FIFO_DD_ONU_SIZE - 1 )
            FIFO_DD.Add( IDLE_BLOCK );
    }
};

#endif // _FSM_DD_H_INCLUDED_

